﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace SetCBLUserImage
{
    public class IntelHexFile
    {
        private const string _HEADER = "**TIFL**";

        public string Header { get; set; }
        public ushort Revision { get; set; }
        public int ObjectType { get; set; }
        public int Unknown1 { get; set; }
        public byte[] DateTime { get; set; }
        public string Name { get; set; }
        public byte DeviceType { get; set; }
        public byte DataType { get; set; }
        public byte[] Data { get; set; }

        private IntelHexFile()
        {
            //Do nothing
        }

        public static IntelHexFile Load(FileStream input, bool isApp)
        {
            return Load(input, isApp, false);
        }

        public static IntelHexFile Load(string fileName, bool isApp)
        {
            return Load(fileName, isApp, false);
        }

        public static IntelHexFile Load(string fileName, bool isApp, bool ignoreChecksums)
        {
            var file = new FileStream(fileName, FileMode.Open);
            var ret = Load(file, isApp, ignoreChecksums);
            file.Close();

            return ret;
        }

        public static IntelHexFile Load(FileStream input, bool isApp, bool ignoreChecksums)
        {
            var start = input.Position;
            var ret = new IntelHexFile();

            //Read the flash header
            var header = new byte[_HEADER.Length];
            input.Read(header, 0, header.Length);
            ret.Header = ASCIIEncoding.ASCII.GetString(header);
            if (ret.Header != _HEADER)
            {
                throw new FileLoadException("Invalid header: " + ret.Header);
            }

            //Read the revision number (BCD)
            var revision = new byte[2];
            input.Read(revision, 0, revision.Length);
            ret.Revision = (ushort)(revision[0] | (revision[1] << 8));

            //Get object type
            ret.ObjectType = input.ReadByte();

            //Get date/time
            ret.DateTime = new byte[4];
            input.Read(ret.DateTime, 0, ret.DateTime.Length);

            ret.Unknown1 = input.ReadByte();

            //Get name
            int length = input.ReadByte();
            var name = new byte[31];
            input.Read(name, 0, name.Length);
            ret.Name = ASCIIEncoding.ASCII.GetString(name, 0, Math.Min(length, name.Length));

            //Get device type
            ret.DeviceType = (byte)input.ReadByte();

            //Get data type
            ret.DataType = (byte)input.ReadByte();

            //Read Intel Hex data
            input.Seek(start + 0x4A, SeekOrigin.Begin);
            var size = new byte[4];
            input.Read(size, 0, size.Length);
            int s = size[0] | (size[1] << 8) | (size[2] << 16) | (size[3] << 24);
            var hexData = new byte[s];
            input.Read(hexData, 0, s);
            var stream = new StringReader(ASCIIEncoding.ASCII.GetString(hexData));
            int extendedAddress = 0;
            var data = new Dictionary<int, byte>();
            while (true)
            {
                var line = stream.ReadLine();
                if (line != null)
                {
                    if (!string.IsNullOrEmpty(line))
                    {
                        //Make sure line starts with colon and has even number of digits
                        if (line[0] != ':' || ((line.Length % 2) != 1))
                        {
                            throw new FileLoadException("Invalid line found: " + line);
                        }

                        //Get the bytes
                        var lineBytes = new byte[(line.Length - 1) / 2];
                        for (int i = 0; i < lineBytes.Length; i++)
                        {
                            lineBytes[i] = (byte)Convert.ToInt32(line.Substring(i * 2 + 1, 2), 16);
                        }

                        //Split into various parts
                        var dataBytes = new byte[lineBytes[0]];
                        var address = (lineBytes[1] << 8) | lineBytes[2];
                        var blockType = lineBytes[3];
                        Array.Copy(lineBytes, 4, dataBytes, 0, dataBytes.Length);

                        //Verify checksum
                        int sum = 0;
                        foreach (var b in lineBytes)
                        {
                            sum = (sum + b) & 0xFF;
                        }

                        if (!ignoreChecksums && sum != 0)
                        {
                            throw new FileLoadException("Invalid checksum for line: " + line);
                        }

                        //Process it
                        switch (blockType)
                        {
                            case 0x00:
                                //Data
                                int baseAddress;
                                if (isApp)
                                {
                                    baseAddress = extendedAddress + address;
                                }
                                else
                                {
                                    baseAddress = (extendedAddress << 4) | address;
                                }

                                for (int i = 0; i < dataBytes.Length; i++)
                                {
                                    data.Add(baseAddress + i, dataBytes[i]);
                                }

                                break;
                            case 0x02:
                                //Extended address record
                                if (isApp)
                                {
                                    extendedAddress = dataBytes[1] * 0x4000;
                                }
                                else
                                {
                                    extendedAddress = (dataBytes[0] << 8) | dataBytes[1];
                                }

                                break;
                            default:
                                //Uh...nothing, I guess
                                break;
                        }
                    }
                }
                else
                {
                    break;
                }
            }

            stream.Close();

            //Build a giant array to hold this data, defaulted to 0xFF
            var min = data.Keys.Min();
            var max = data.Keys.Max();
            ret.Data = new byte[max - min + 1];
            for (int i = 0; i < ret.Data.Length; i++)
            {
                ret.Data[i] = 0xFF;
            }

            //Fill it
            foreach (var b in data)
            {
                ret.Data[b.Key - min] = b.Value;
            }

            return ret;
        }

        public byte[] GetSaveData()
        {
            var stream = new MemoryStream();

            //Write flash header
            stream.Write(ASCIIEncoding.ASCII.GetBytes(Header), 0, Header.Length);

            //Write revision number
            var revision = new byte[2];
            revision[0] = (byte)(Revision & 0xFF);
            revision[1] = (byte)((Revision >> 8) & 0xFF);
            stream.Write(revision, 0, revision.Length);

            //Write object type
            stream.Write(new byte[] { (byte)ObjectType }, 0, 1);

            //Write date/time
            stream.Write(DateTime, 0, DateTime.Length);

            stream.Write(new byte[] { (byte)Unknown1 }, 0, 1);

            //Write name
            var n = new byte[32];
            n[0] = (byte)Name.Length;
            ASCIIEncoding.ASCII.GetBytes(Name, 0, Name.Length, n, 1);
            stream.Write(n, 0, n.Length);

            //Write device type
            stream.Write(new byte[] { (byte)DeviceType }, 0, 1);

            //Write data type
            stream.Write(new byte[] { (byte)DataType }, 0, 1);

            //Generate the Intel Hex for the data
            var intel = new StringWriter();
            int offset = 0;
            while (offset < Data.Length)
            {
                if ((offset & 0xFFFF) == 0)
                {
                    //Write address line
                    int address = offset >> 4;
                    intel.WriteLine(_GetIntelHexLine(0x0000, 0x02,
                        new byte[] { (byte)((address >> 8) & 0xFF), (byte)(address & 0xFF) }));
                }

                //Write data
                var data = new byte[Math.Min(0x10, Data.Length - offset)];
                Array.Copy(Data, offset, data, 0, data.Length);
                intel.WriteLine(_GetIntelHexLine(offset & 0xFFFF, 0x00, data));

                offset += data.Length;
            }

            //Write end-of-file record
            intel.WriteLine(_GetIntelHexLine(0, 0x01, null));

            //Write size and data as Intel Hex
            var hex = intel.ToString();
            stream.Seek(0x4A, SeekOrigin.Begin);
            var size = new byte[4];
            size[0] = (byte)(hex.Length & 0xFF);
            size[1] = (byte)((hex.Length >> 8) & 0xFF);
            size[2] = (byte)((hex.Length >> 16) & 0xFF);
            size[3] = (byte)((hex.Length >> 24) & 0xFF);
            stream.Write(size, 0, size.Length);
            var bytes = ASCIIEncoding.ASCII.GetBytes(hex);
            stream.Write(bytes, 0, bytes.Length);

            return stream.ToArray();
        }

        private string _GetIntelHexLine(int address, byte type, byte[] data)
        {
            var ret = ":";

            int length = data != null ? data.Length : 0;
            int checksum = length + (address & 0xFF) + ((address >> 8) & 0xFF) + type;
            ret += length.ToString("X02");
            ret += address.ToString("X04");
            ret += type.ToString("X02");
            if (data != null)
            {
                foreach (var b in data)
                {
                    ret += b.ToString("X02");
                    checksum += b;
                }
            }

            ret += ((~checksum + 1) & 0xFF).ToString("X02");

            return ret;
        }
    }
}
