﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace SetCBLUserImage
{
    class Startup
    {
        const string _STRIP_MODELS = "73,82,83,83P,86,89,92,92P";

        static void Main(string[] args)
        {
            //Get command-line arguments:
            const string HELP_PREFIX = "--help";
            const string INPUT_PREFIX = "--input=";
            const string OUTPUT_PREFIX = "--output=";
            const string IGNORE_CHECKSUMS_PREFIX = "--ignore-checksums";
            const string STRIP_FILES = "--strip-files-for=";
            const string EXCLUDE_OS = "--exclude-os";
            var files = new List<string>();
            string input = string.Empty;
            string output = string.Empty;
            bool ignoreChecksums = false;
            bool excludeOS = false;
            var stripFrom = new List<string>();

            try
            {
                foreach (var arg in args)
                {
                    if (arg.StartsWith(INPUT_PREFIX, StringComparison.CurrentCultureIgnoreCase))
                    {
                        //  Input filename
                        input = arg.Substring(INPUT_PREFIX.Length);
                        Console.WriteLine("Specified input file: " + input);
                    }
                    else if (arg.StartsWith(OUTPUT_PREFIX, StringComparison.CurrentCultureIgnoreCase))
                    {
                        //  Output filename
                        output = arg.Substring(OUTPUT_PREFIX.Length);
                        Console.WriteLine("Specified output file: " + output);
                    }
                    else if (arg.StartsWith(HELP_PREFIX, StringComparison.CurrentCultureIgnoreCase))
                    {
                        _DisplayUsage();
                        return;
                    }
                    else if (arg.StartsWith(IGNORE_CHECKSUMS_PREFIX, StringComparison.CurrentCultureIgnoreCase))
                    {
                        ignoreChecksums = true;
                        Console.WriteLine("Specified ignoring checksums");
                    }
                    else if (arg.StartsWith(EXCLUDE_OS, StringComparison.CurrentCultureIgnoreCase))
                    {
                        excludeOS = true;
                        Console.WriteLine("Specified excluding OS update");
                    }
                    else if (arg.StartsWith(STRIP_FILES, StringComparison.CurrentCultureIgnoreCase))
                    {
                        var valid = _STRIP_MODELS.Split(new char[] { ',' }).Select(t => t.ToUpper());
                        var models = arg.Substring(STRIP_FILES.Length).ToUpper();
                        foreach (var model in models.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries))
                        {
                            if (valid.Contains(model))
                            {
                                stripFrom.Add(model);

                                Console.WriteLine("Specified removing files for " + model);
                            }
                        }
                    }
                    else
                    {
                        //  List of files to embed
                        if (Directory.Exists(arg))
                        {
                            Console.WriteLine("Specified directory:\t" + arg);

                            foreach (var file in Directory.GetFiles(arg))
                            {
                                var f = Path.Combine(file, arg);
                                Console.WriteLine("\tSpecified file:\t" + f);

                                files.Add(f);
                            }
                        }
                        else
                        {
                            Console.WriteLine("Specified file:\t" + arg);

                            files.Add(arg);
                        }
                    }
                }

                //Verify we have what we need to continue
                if (string.IsNullOrEmpty(input))
                {
                    throw new ArgumentException("Input file name not specified");
                }
                else if (string.IsNullOrEmpty(output))
                {
                    throw new ArgumentException("Output file name not specified");
                }

                //Read input file into memory
                Console.WriteLine("Loading input file...");
                var updateFile = CBL2OSUpdate.Load(input, ignoreChecksums);

                //Remove files for any models specified
                foreach (var model in stripFrom)
                {
                    _RemoveFiles(updateFile, model);
                }

                foreach (var file in files)
                {
                    //For each file:
                    //  Read it into memory
                    var image = UserImageRecord.LoadFromFile(file, ignoreChecksums);

                    if (image != null)
                    {
                        //  Delete it if it exists
                        var matches = updateFile.UserImage.Files.Where(t => t.Type == image.Type &&
                            t.ToString() == image.ToString()).ToList();
                        if (matches.Any())
                        {
                            foreach (var match in matches)
                            {
                                updateFile.UserImage.Files.Remove(match);
                            }
                        }

                        //  Add it
                        updateFile.UserImage.Files.Add(image);
                    }
                    else
                    {
                        Console.WriteLine(string.Format("WARNING: '{0}' is not a valid calculator file.", file));
                    }
                }

                //Save output file
                Console.WriteLine("Writing output file...");
                updateFile.Save(output, excludeOS);

                Console.WriteLine("Done.");
            }
            catch (Exception ex)
            {
                _DisplayUsage();

                Console.WriteLine("ERROR: " + ex.ToString());

                Environment.ExitCode = -1;
                return;
            }
        }

        private static void _RemoveFiles(CBL2OSUpdate updateFile, string model)
        {
            byte machineID;

            //First figure out the machine ID from the model
            switch (model)
            {
                case "73":
                    machineID = 0x74;
                    break;
                case "82":
                    machineID = 0x82;
                    break;
                case "83":
                    machineID = 0x83;
                    break;
                case "83P":
                    machineID = 0x73;
                    break;
                case "86":
                    machineID = 0x86;
                    break;
                case "89":
                    machineID = 0x98;
                    break;
                case "92":
                    machineID = 0x89;
                    break;
                case "92P":
                    machineID = 0x88;
                    break;
                default:
                    throw new InvalidOperationException("Invalid model specified: " + model);
            }

            var matches = updateFile.UserImage.Files.Where((t) =>
                {
                    bool ret = false;

                    if (t.Type == UserImageRecord.RecordType.Application)
                    {
                        var app = t as ApplicationRecord;
                        ret = (machineID == app.MachineID);
                    }
                    else if (t.Type == UserImageRecord.RecordType.Program)
                    {
                        var prog = t as ProgramRecord;
                        ret = (machineID == prog.MachineID);
                    }

                    return ret;
                });
            if (matches.Any())
            {
                foreach (var match in matches)
                {
                    updateFile.UserImage.Files.Remove(match);
                }
            }
        }

        private static void _DisplayUsage()
        {
            var name = Assembly.GetExecutingAssembly().GetName();
            Console.WriteLine(name.Name + " " + name.Version.ToString());
            Console.WriteLine("Allows manipulating the \"user image\" included with TI-CBL2/LabPro updates.");
            Console.WriteLine();
            Console.WriteLine("Arguments:");
            Console.WriteLine("--help\t\t\tThis help text.");
            Console.WriteLine("--input\t\t\tInput C2U OS upgrade file name.");
            Console.WriteLine("--output\t\tOutput C2U OS upgrade file name.");
            Console.WriteLine("--ignore-checksums\tSkip checksum checks on Intel Hex and user image data.");
            Console.WriteLine("--strip-files-for\tComma-delimited list of models to remove all files for.");
            Console.WriteLine("\tValid models: 73,82,83,83P,86,89,92,92P");
            Console.WriteLine();
            Console.WriteLine("Specify a list of files to include with the update on the command line.");
            Console.WriteLine("Example: --input=input.c2u --output=output.c2u MYPRGM.8XP MYAPP.8XK");
            Console.WriteLine();
        }
    }
}
